import re
from dataclasses import dataclass, field
from typing import Any, Optional

_split_reg = re.compile(r'\s?(-\w|--\w+)[=\s]')


@dataclass
class CommandArgs:
    default: Any
    help: Optional[str] = None
    keyword_args: tuple[str, ...] = field(default_factory=tuple)

    name: Optional[str] = field(init=False, default=None)
    data_type: Any = field(init=False, default=Any)

    # @property
    # def name(self):
    #     return self._name


def CmdArg(
        *args: str,
        help: str = None
):
    default = ... if args[0] is ... else None
    keyword_args = args if default is None else tuple()
    return CommandArgs(
        default=default,
        help=help,
        keyword_args=keyword_args
    )


CommandParams = list[CommandArgs]


def _get_cmd_args(cmd: str):
    positional_args = []
    keywork_params = {}

    is_positional_arg = True
    last_position = 0

    cmd_split = _split_reg.split(cmd)
    for i, _arg in enumerate(cmd_split):
        if is_positional_arg and _arg.startswith('-'):
            is_positional_arg = False

        if is_positional_arg:
            positional_args.append(_arg)
        elif i > last_position:
            last_position = i + 1
            keywork_params[_arg] = cmd_split[last_position]

    return positional_args, keywork_params


class PosParamsCountError(Exception):
    pass


class ParamNotFoundError(Exception):
    pass


def parse_args(cmd_args: list[str],
               positional_params: CommandParams,
               keyword_params: CommandParams) -> dict:
    positional_args, keyword_args = _get_cmd_args(' '.join(cmd_args))

    # Positional arguments

    if len(positional_params) != len(positional_args):
        raise PosParamsCountError(f'Expected {len(positional_params)} positional values but got {len(positional_args)}')
    fn_args = {
        _param.name: _param_value
        for _param, _param_value in zip(positional_params, positional_args)
    }

    # Keyword Arguments

    _keyword_mapping = {
        _keyword_arg: cmd.name
        for cmd in keyword_params
        for _keyword_arg in cmd.keyword_args
    }

    for _keyword_arg_key, _keyword_arg_value in keyword_args.items():
        _keyword_param = _keyword_mapping.get(_keyword_arg_key)
        if not _keyword_param:
            raise ParamNotFoundError(f'Could not found param {_keyword_arg_key}')
        fn_args[_keyword_param] = _keyword_arg_value

    return fn_args



import abc
from dataclasses import dataclass

from ..command import Command, Option


@dataclass
class Formatter(abc.ABC):

    @abc.abstractmethod
    def print_help(self,
                   commands: dict[str, Command],
                   options: list[Option]) -> None:
        ...

    @abc.abstractmethod
    def print_cmd_error(self, cmd: str) -> None:
        ...



import sys
from dataclasses import dataclass, field
from functools import wraps
from typing import get_type_hints, List, Any

from .command import Command, Option
from .utils import parse_args, CommandArgs
from .formatter import Formatter, RichFormatter


@dataclass
class Parser:
    description: str

    formatter: Formatter = RichFormatter()
    _options: List[Option] = field(init=False, default_factory=list)
    _commands: dict[str, Command] = field(init=False, default_factory=dict)

    @property
    def commands(self):
        return self._commands

    def print_help(self):
        self.formatter.print_help(
            commands=self.commands,
            options=self._options
        )

    def run(self):
        try:
            _, cmd, *args = sys.argv
        except ValueError:
            return

        _command = self._commands.get(cmd)
        if not _command:
            self.formatter.print_cmd_error(cmd)
            return
        _args_dict = parse_args(args,
                                _command.positional_params,
                                _command.keyword_params)
        _command.run(**_args_dict)

    def command(self, cmd: str, help: str = None):
        def _command(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                return f(*args, **kwargs)

            positional_params, optional_params = [], []
            defaults: List[CommandArgs] = f.__defaults__ or []
            for (param_name, param_type), cmd_args in zip(get_type_hints(f).items(),
                                                          defaults):
                cmd_args.name = param_name
                cmd_args.data_type = param_type
                if cmd_args.default is ...:
                    positional_params.append(cmd_args)
                else:
                    optional_params.append(cmd_args)

            self._commands[cmd] = Command(name=cmd,
                                          fn=wrapper,
                                          positional_params=positional_params,
                                          keyword_params=optional_params,
                                          help=help)
            return wrapper

        return _command

    def add_argument(self, *args: str, help: str = None):
        self._options.append(Option(help, args=args))



from dataclasses import dataclass, field
from rich import box
from rich.console import Console
from rich.table import Table

from ..command import Command, Option
from .base import Formatter


@dataclass
class RichFormatter(Formatter):

    _console: Console = field(init=False, default=Console(markup=True))

    def print_help(self,
                   commands: dict[str, Command],
                   options: list[Option]):

        _help = '[bold bright_white]GLOBAL OPTIONS[/bold bright_white]'
        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column('Command name', style="cyan", width=15)
        table.add_column('Command help')
        for _command in options:
            if len(_command.args) == 0:
                name = None
            else:
                first_arg, *other_args = _command.args
                if other_args:
                    other_args = ', '.join(other_args)
                    name = f'{first_arg} [bright_white]({other_args})[/bright_white]'
                else:
                    name = first_arg
            table.add_row(name, _command.help)
        self._console.print(_help, table)

        _help = '[bold bright_white]AVAILABLE COMMANDS[/bold bright_white]'
        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column('Command name', style="cyan", width=15)
        table.add_column('Command help')
        for _command in commands.values():
            table.add_row(_command.name, _command.help)
        self._console.print(_help, table)

    def print_cmd_error(self, cmd: str):
        self._console.print(f'[red]Unknown command "[bold]{cmd}[/bold]"[/red]')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piou', 'piou.formatter']

package_data = \
{'': ['*']}

install_requires = \
['rich>=10.11.0,<11.0.0']

extras_require = \
{'pydantic': ['pydantic>=1.8.2,<2.0.0']}

setup_kwargs = {
    'name': 'piou',
    'version': '0.1.2',
    'description': 'A CLI tool',
    'long_description': "# Piou  \n\n\nA CLI tool to build beautiful command-line interfaces with type validation.\n\nIt is as simple as\n\n```python\nfrom piou import Parser, CmdArg\n\nparser = Parser(description='A CLI tool')\n\nparser.add_argument('-h', '--help', help='Display this help message')\nparser.add_argument('-q', '--quiet', help='Do not output any message')\nparser.add_argument('--verbose', help='Increase verbosity')\n\n\n@parser.command(cmd='foo',\n                help='Run foo command')\ndef foo_main(\n    foo1: int = CmdArg(..., help='Foo arguments'),\n    foo2: str = CmdArg('-f', '--foo2', help='Foo arguments')\n):\n    print('Ran foo main with ', foo1)\n    print('foo2: ', foo2)\n\n\n@parser.command(cmd='bar', help='Run bar command')\ndef bar_main():\n    pass\n\n\nif __name__ == '__main__':\n    parser.print_help()\n    parser.run()\n```\nThe output will look like this: \n\n![example](https://github.com/Andarius/piou/blob/master/docs/example.png?raw=true)",
    'author': 'Julien Brayere',
    'author_email': 'julien.brayere@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andarius/pioupiou',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
from typing import Callable, Optional

from .utils import CommandParams


@dataclass
class Command:
    name: str
    help: Optional[str]
    fn: Callable
    positional_params: CommandParams = field(default_factory=list)
    keyword_params: CommandParams = field(default_factory=list)

    @property
    def params(self) -> CommandParams:
        return [
            *self.positional_params,
            *self.keyword_params
        ]

    def run(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __post_init__(self):
        if not self.keyword_params:
            return

        _keyword_args = set()
        for _param in self.keyword_params:
            for _keyword_arg in _param.keyword_args:
                if _keyword_arg in _keyword_args:
                    raise ValueError(f'Duplicate keyword args found "{_keyword_arg}"')
                _keyword_args.add(_keyword_arg)


@dataclass
class Option:
    help: Optional[str]
    args: tuple[str, ...]
